/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: pyauthconn.c,v 1.2 2000/08/08 14:39:00 bazsi Exp $
 *
 ***************************************************************************/

#include <zorp/zpython.h>
#include <zorp/log.h>

extern PyTypeObject z_py_zorp_authconn_type;

static PyObject *
z_py_zorp_authconn_new_instance(PyObject *s, PyObject *args)
{
  ZorpSockAddr *sa;
  ZorpAuthConnection *self;
  char *peername;
  char buf[128];
  
  if (!PyArg_ParseTuple(args, "Os", &sa, &peername))
    return NULL;
  if (!z_py_zorp_sockaddr_check(sa))
    {
      PyErr_SetString(PyExc_TypeError, "ZorpSockAddr expected");
      return NULL;
    }
  z_message(5, "Connecting to ZAS Auth server: %s, peer=%s", z_sockaddr_format(sa->sa, buf, sizeof(buf)), peername);
  self = PyObject_NEW(ZorpAuthConnection, &z_py_zorp_authconn_type);
  self->authconn = z_authconn_new_ssl(sa->sa, peername);
  if (!self->authconn)
    {
      PyErr_SetString(PyExc_IOError, "Error establishing auth connection");
      Py_XDECREF(self);
      return NULL;
    }
  return (PyObject *) self;
}

static void
z_py_zorp_authconn_free(ZorpAuthConnection *self)
{
  z_message(5, "Dropping connection to ZAS Auth server");
  if (self->authconn)
    {
      z_authconn_destroy(self->authconn);
      self->authconn = NULL;
    }
  PyMem_DEL(self);
}

static PyObject *
z_py_zorp_authconn_getattr(ZorpAuthConnection *self, char *name)
{
}

static int
z_py_zorp_authconn_setattr(ZorpAuthConnection *self, char *name, PyObject *value)
{
}

PyTypeObject z_py_zorp_authconn_type =
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,                                        /*ob_size*/
  "ZorpAuthConnection",                     /*tp_name*/
  sizeof(ZorpAuthConnection),               /*tp_basicsize*/
  0,                                        /*tp_itemsize*/
  /* methods */
  (destructor)z_py_zorp_authconn_free,      /*tp_dealloc*/
  (printfunc)0,                             /*tp_print*/
  (getattrfunc)z_py_zorp_authconn_getattr,  /*tp_getattr*/
  (setattrfunc)z_py_zorp_authconn_setattr,  /*tp_setattr*/
  (cmpfunc)0,                               /*tp_compare*/
  (reprfunc)0,                              /*tp_repr*/
  0,                                        /*tp_as_number*/
  0,                   		            /*tp_as_sequence*/
  0,                   		            /*tp_as_mapping*/
  (hashfunc)0,         		            /*tp_hash*/
  (ternaryfunc)0,      		            /*tp_call*/
  (reprfunc)0,         		            /*tp_str*/

  /* Space for future expansion */
  0L,0L,0L,0L,
  "ZorpAuthConnection class", /* Documentation string */
};

PyMethodDef z_py_zorp_authconn_funcs[] =
{
  { "AuthConnection",  z_py_zorp_authconn_new_instance, METH_VARARGS },
  { NULL,      NULL }   /* sentinel*/
};


void
z_py_zorp_authconn_init(void)
{
  Py_InitModule("Zorp.ZAS", z_py_zorp_authconn_funcs);
}
