/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: stream.c,v 1.9 2000/08/01 10:10:36 bazsi Exp $
 *
 ***************************************************************************/

#include <zorp/stream.h>
#include <zorp/log.h>

#include <string.h>
#include <errno.h>

typedef ZStream ZRealStream;

static GIOError
z_stream_read_method(GIOChannel *channel,
		     gchar *buf,
		     guint count,
		     guint *bytes_read)
{
  ZRealStream *self = (ZRealStream *) channel;
  gint res;
  /* FIXME: todo, datadump */
  res = g_io_channel_read(self->channel, buf, count, bytes_read);
  z_debug(10, "%s: reading channel, fd=%d, count=%d", self->name, self->fd, *bytes_read);
  return res;
}

static GIOError
z_stream_write_method(GIOChannel *channel,
		      gchar *buf,
		      guint count,
		      guint *bytes_written)
{
  ZRealStream *self = (ZRealStream *) channel;

  /* FIXME: todo, datadump */
  z_debug(10, "%s: writing channel, fd=%d, count=%d", self->name, self->fd, count);
  return g_io_channel_write(self->channel, buf, count, bytes_written);
}

static void
z_stream_close_method(GIOChannel *channel)
{
  ZRealStream *self = (ZRealStream *) channel;

  z_debug(10, "%s: closing channel, fd=%d", self->name, self->fd);
  g_io_channel_close(self->channel);
}

static void
z_stream_free(GIOChannel *channel)
{
  ZRealStream *self = (ZRealStream *) channel;

  g_io_channel_unref(self->channel);
  g_free(channel);
}

GIOFuncs z_stream_funcs =
{
  z_stream_read_method,
  z_stream_write_method,
  NULL,
  z_stream_close_method,
  NULL,
  z_stream_free
};

/*+

  Allocate and initialize a ZThread instance with the given fd and name.

  Parameters:
    fd        fd to wrap this stream around
    name      name to identify to stream in logs
  
  Returns:
    ZStream instance

  +*/
ZStream *
z_stream_new(gint fd, gchar *name)
{
  ZRealStream *self = g_new0(ZRealStream, 1);

  g_io_channel_init(&self->super);
  self->fd = fd;
  self->super.funcs = &z_stream_funcs;

  self->channel = g_io_channel_unix_new(fd);
  strncpy(self->name, name, sizeof(self->name) - 1);
  return (ZStream *) self;
}

