/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ftphash.c,v 1.8 2000/08/10 13:02:17 bazsi Exp $
 *
 * Author:  Andras Kis-Szabo <kisza@sch.bme.hu>
 * Auditor:
 * Notes:
 *
 ***************************************************************************/

#include <zorp/zorp.h>
#include <zorp/stream.h>
#include <zorp/readline.h>
#include <zorp/proxy.h>
#include <zorp/policy.h>
#include <zorp/thread.h>
#include <zorp/zpython.h>
#include <zorp/log.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "ftp.h"
#include "ftphash.h"
#include "ftpmain.h"
#include "ftpparse.h"

//#define ENABLE_DEBUG

// APASV EPRT EPSV LPRT LPSV .... ???
static struct _FTPCommandDescriptor ftp_commands[] = {
  {
   "ABOR", active_flag_ABOR, 1, ftp_command_parse_noarg, "sendABOR", 0}
  ,
  {
   "ACCT", active_flag_ACCT, 2, ftp_command_parse_ACCT, "userAUTH", 0}
  ,
  {
   "ADAT", active_flag_ADAT, 3, ftp_command_parse_sftp, "generalCommand",
   0}
  ,
  {
   "ALLO", active_flag_ALLO, 4, ftp_command_parse_ALLO, "miscCommand", 0}
  ,
  {
   "APPE", active_flag_APPE, 5, ftp_command_parse_path, "transferCommand",
   0}
  ,
  {
   "APSV", active_flag_APSV, 6, ftp_command_parse_PASV, "generalCommand",
   0}
  ,
  {
   "AUTH", active_flag_AUTH, 7, ftp_command_parse_sftp, "generalCommand",
   0}
  ,
  {
   "CCC", active_flag_CCC, 8, ftp_command_parse_sftp, "generalCommand", 0}
  ,
  {
   "CDUP", active_flag_CDUP, 9, ftp_command_parse_noarg,
   "directoryChange", 0}
  ,
  {
   "CONF", active_flag_CONF, 10, ftp_command_parse_sftp, "generalCommand",
   0}
  ,
  {
   "CWD", active_flag_CWD, 11, ftp_command_parse_path, "directoryChange",
   0}
  ,
  {
   "DELE", active_flag_DELE, 12, ftp_command_parse_path,
   "directoryManipulation", 0}
  ,
  {
   "ENC", active_flag_ENC, 13, ftp_command_parse_sftp, "generalCommand", 0}
  ,
  {
   "EPRT", active_flag_EPRT, 14, ftp_command_parse_string, "generalCommand",
   0}
  ,
  {
   "EPSV", active_flag_EPSV, 15, ftp_command_parse_string, "generalCommand",
   0}
  ,
  {
   "FEAT", active_flag_FEAT, 16, ftp_command_parse_noarg, "miscCommand", 0}
  ,
  {
   "HELP", active_flag_HELP, 17, ftp_command_parse_HELP, "miscCommand", 0}
  ,
  {
   "LIST", active_flag_LIST, 18, ftp_command_parse_path,
   "transferCommand", 1}
  ,
  {
   "LPRT", active_flag_LPRT, 19, ftp_command_parse_string, "generalCommand",
   0}
  ,
  {
   "LPSV", active_flag_LPSV, 20, ftp_command_parse_string, "generalCommand",
   0}
  ,
  {
   "MAIL", active_flag_MAIL, 21, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "MDTM", active_flag_MDTM, 22, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "MIC", active_flag_MIC, 23, ftp_command_parse_sftp, "generalCommand", 0}
  ,
  {
   "MKD", active_flag_MKD, 24, ftp_command_parse_path,
   "directoryManipulation", 0}
  ,
  {
   "MLFL", active_flag_MLFL, 25, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "MODE", active_flag_MODE, 26, ftp_command_parse_MODE, "generalCommand",
   0}
  ,
  {
   "MRCP", active_flag_MRCP, 27, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "MRSQ", active_flag_MRSQ, 28, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "MSAM", active_flag_MSAM, 29, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "MSND", active_flag_MSND, 20, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "MSOM", active_flag_MSOM, 21, ftp_command_parse_string, "mailCommand",
   0}
  ,
  {
   "NLST", active_flag_NLST, 22, ftp_command_parse_path,
   "transferCommand", 1}
  ,
  {
   "NOOP", active_flag_NOOP, 23, ftp_command_parse_noarg,
   "generalCommand", 0}
  ,
  {
   "OPTS", active_flag_OPTS, 24, ftp_command_parse_noarg, "miscCommand", 0}
  ,
  {
   "PASS", active_flag_PASS, 25, ftp_command_parse_PASS, "password", 0}
  ,
  {
   "PASV", active_flag_PASV, 26, ftp_command_parse_PASV, "generalCommand",
   0}
  ,
  {
   "PBSZ", active_flag_PBSZ, 27, ftp_command_parse_sftp, "generalCommand",
   0}
  ,
  {
   "PORT", active_flag_PORT, 28, ftp_command_parse_PORT, "portCommand", 0}
  ,
  {
   "PROT", active_flag_PROT, 29, ftp_command_parse_sftp, "generalCommand",
   0}
  ,
  {
   "PWD", active_flag_PWD, 30, ftp_command_parse_noarg, "generalCommand",
   0}
  ,
  {
   "QUIT", active_flag_QUIT, 31, ftp_command_parse_QUIT, "quit", 0}
  ,
  {
   "REIN", active_flag_REIN, 32, ftp_command_parse_REIN, "generalCommand",
   0}
  ,
  {
   "REST", active_flag_REST, 33, ftp_command_parse_marker, "miscCommand",
   0}
  ,
  {
   "RETR", active_flag_RETR, 34, ftp_command_parse_path,
   "transferCommand", 1}
  ,
  {
   "RMD", active_flag_RMD, 35, ftp_command_parse_path,
   "directoryManipulation", 0}
  ,
  {
   "RNFR", active_flag_RNFR, 36, ftp_command_parse_path,
   "directoryManipulation", 0}
  ,
  {
   "RNTO", active_flag_RNTO, 37, ftp_command_parse_path,
   "directoryManipulation", 0}
  ,
  {
   "SITE", active_flag_SITE, 38, ftp_command_parse_string, "siteCommand",
   0}
  ,
  {
   "SMNT", active_flag_SMNT, 30, ftp_command_parse_path, "miscCommand", 0}
  ,
  {
   "STAT", active_flag_STAT, 31, ftp_command_parse_path, "miscCommand", 0}
  ,
  {
   "STOR", active_flag_STOR, 32, ftp_command_parse_path,
   "transferCommand", 1}
  ,
  {
   "STOU", active_flag_STOU, 33, ftp_command_parse_path,
   "transferCommand", 1}
  ,
  {
   "STRU", active_flag_STRU, 34, ftp_command_parse_STRU, "generalCommand",
   0}
  ,
  {
   "SYST", active_flag_SYST, 35, ftp_command_parse_noarg, "miscCommand", 0}
  ,
  {
   "TYPE", active_flag_TYPE, 36, ftp_command_parse_TYPE, "generalCommand",
   0}
  ,
  {
   "USER", active_flag_USER, 37, ftp_command_parse_USER, "user", 0}
  ,
  {
   "XCUP", active_flag_XCUP, 38, ftp_command_parse_noarg,
   "directoryChange", 0}
  ,
  {
   "XCWD", active_flag_XCWD, 39, ftp_command_parse_path,
   "directoryChange", 0}
  ,
  {
   "XMKD", active_flag_XMKD, 30, ftp_command_parse_path,
   "directoryManipulation", 0}
  ,
  {
   "XPWD", active_flag_XPWD, 31, ftp_command_parse_noarg,
   "generalCommand", 0}
  ,
  {
   "XRMD", active_flag_XRMD, 32, ftp_command_parse_path,
   "directoryManipulation", 0}
  ,
  {
   NULL, 0, 0, NULL, NULL, 0}
};
  //  ftp_command_hash_add ("SIZE", active_flag_, 39, ftp_command_parse_SIZE,0);

static struct _FTPInternalCommand ftp_internal_commands[] = {
  {"ABOR", 0},
  {"ACCT", 0},
  {"ADAT", 0},
  {"ALLO", 0},
  {"APPE", 0},
  {"APSV", 0},
  {"AUTH", 0},
  {"CCC", 0},
  {"CDUP", 0},
  {"CONF", 0},
  {"CWD", 0},
  {"DELE", 0},
  {"ENC", 0},
  {"EPRT", 0},
  {"EPSV", 0},
  {"FEAT", 0},
  {"HELP", 0},
  {"LIST", 0},
  {"LPRT", 0},
  {"LPSV", 0},
  {"MAIL", 0},
  {"MDTM", 0},
  {"MIC", 0},
  {"MKD", 0},
  {"MLFL", 0},
  {"MODE", 0},
  {"MRCP", 0},
  {"MRSQ", 0},
  {"MSAM", 0},
  {"MSND", 0},
  {"MSOM", 0},
  {"NLST", 0},
  {"NOOP", 0},
  {"OPTS", 0},
  {"PASS", 0},
  {"PASV", 0},
  {"PBSZ", 0},
  {"PORT", 0},
  {"PROT", 0},
  {"PWD", 0},
  {"QUIT", 0},
  {"REIN", 0},
  {"REST", 0},
  {"RETR", 0},
  {"RMD", 0},
  {"RNFR", 0},
  {"RNTO", 0},
  {"SITE", 0},
  {"STAT", 0},
  {"STOR", 0},
  {"STOU", 0},
  {"STRU", 0},
  {"SYST", 0},
  {"TYPE", 0},
  {"USER", 0},
  {"XCUP", 0},
  {"XCWD", 0},
  {"XMKD", 0},
  {"XPWD", 0},
  {"XRMD", 0},
  {NULL, 0}
};
  //  ftp_local_hash_add (self, "SIZE", 1);

static struct _FTPAnswerDescriptor ftp_answers[] = {
//define FTP_STATE_CONNECT           0
  {"00120", 1},
  {"00220", 1},
  {"00421", 1},
//define FTP_STATE_LOGIN             1
  {"01211HELP", 1},
  {"01214HELP", 1},
  {"01500HELP", 1},
  {"01501HELP", 1},
  {"01502HELP", 1},
  {"01421HELP", 1},
  // WU-FTPD using this ...
  {"01530HELP", 1},
//define FTP_STATE_LOGINAUTH         3
  {"03211HELP", 1},
  {"03214HELP", 1},
  {"03500HELP", 1},
  {"03501HELP", 1},
  {"03502HELP", 1},
  {"03421HELP", 1},
//define FTP_STATE_LOGIN_U           11
  {"11230USER", 1},
  {"11232USER", 1},
  {"11530USER", 1},
  {"11500USER", 1},
  {"11501USER", 1},
  {"11421USER", 1},
  {"11331USER", 1},
  {"11332USER", 1},
  {"11336USER", 1},
//define FTP_STATE_LOGIN_P           12
  {"12230PASS", 1},
  {"12202PASS", 1},
  {"12530PASS", 1},
  {"12500PASS", 1},
  {"12501PASS", 1},
  {"12503PASS", 1},
  {"12421PASS", 1},
  {"12332PASS", 1},
//define FTP_STATE_LOGIN_A           13
  {"13230ACCT", 1},
  {"13202ACCT", 1},
  {"13530ACCT", 1},
  {"13500ACCT", 1},
  {"13501ACCT", 1},
  {"13503ACCT", 1},
  {"13421ACCT", 1},
//define FTP_STATE_QUIT              5
  {"05221QUIT", 1},
//define FTP_STATE_CONVERSATION      4
  {"04225ABOR", 1},
  {"04226ABOR", 1},
  {"04500ABOR", 1},
  {"04501ABOR", 1},
  {"04502ABOR", 1},
  {"04421ABOR", 1},

  {"04230ACCT", 1},
  {"04202ACCT", 1},
  {"04530ACCT", 1},
  {"04500ACCT", 1},
  {"04501ACCT", 1},
  {"04503ACCT", 1},
  {"04421ACCT", 1},

  {"04235ADAT", 1},
  {"04335ADAT", 1},
  {"04500ADAT", 1},
  {"04501ADAT", 1},
  {"04503ADAT", 1},
  {"04535ADAT", 1},
  {"04421ADAT", 1},

  {"04200ALLO", 1},
  {"04202ALLO", 1},
  {"04500ALLO", 1},
  {"04501ALLO", 1},
  {"04504ALLO", 1},
  {"04421ALLO", 1},
  {"04530ALLO", 1},

  {"04125APPE", 1},
  {"04150APPE", 1},
  {"04110APPE", 1},
  {"04226APPE", 1},
  {"04250APPE", 1},
  {"04425APPE", 1},
  {"04426APPE", 1},
  {"04451APPE", 1},
  {"04551APPE", 1},
  {"04552APPE", 1},
  {"04532APPE", 1},
  {"04450APPE", 1},
  {"04550APPE", 1},
  {"04452APPE", 1},
  {"04553APPE", 1},
  {"04500APPE", 1},
  {"04501APPE", 1},
  {"04502APPE", 1},
  {"04421APPE", 1},
  {"04530APPE", 1},
  {"04534APPE", 1},
  {"04535APPE", 1},

  {"04151APSV", 1},
  {"04227APSV", 1},
  {"04500APSV", 1},
  {"04501APSV", 1},
  {"04502APSV", 1},
  {"04421APSV", 1},
  {"04530APSV", 1},

  {"04234AUTH", 1},
  {"04334AUTH", 1},
  {"04502AUTH", 1},
  {"04504AUTH", 1},
  {"04534AUTH", 1},
  {"04431AUTH", 1},
  {"04500AUTH", 1},
  {"04501AUTH", 1},
  {"04421AUTH", 1},

  //ftp_answer_hash_add ("04CCC");

  {"04200CDUP", 1},
  {"04500CDUP", 1},
  {"04501CDUP", 1},
  {"04502CDUP", 1},
  {"04421CDUP", 1},
  {"04530CDUP", 1},
  {"04550CDUP", 1},

  {"04535CONF", 1},
  {"04533CONF", 1},
  {"04500CONF", 1},
  {"04501CONF", 1},
  {"04421CONF", 1},

  {"04250CWD", 1},
  {"04500CWD", 1},
  {"04501CWD", 1},
  {"04502CWD", 1},
  {"04421CWD", 1},
  {"04530CWD", 1},
  {"04550CWD", 1},

  {"04250DELE", 1},
  {"04450DELE", 1},
  {"04550DELE", 1},
  {"04500DELE", 1},
  {"04501DELE", 1},
  {"04502DELE", 1},
  {"04421DELE", 1},
  {"04530DELE", 1},

  {"04535ENC", 1},
  {"04533ENC", 1},
  {"04500ENC", 1},
  {"04501ENC", 1},
  {"04421ENC", 1},

  //ftp_answer_hash_add ("04EPRT");

  //ftp_answer_hash_add ("04EPSV");

  {"04500FEAT", 1},

  {"04211HELP", 1},
  {"04214HELP", 1},
  {"04500HELP", 1},
  {"04501HELP", 1},
  {"04502HELP", 1},
  {"04421HELP", 1},

  {"04125LIST", 1},
  {"04150LIST", 1},
  {"04226LIST", 1},
  {"04250LIST", 1},
  {"04425LIST", 1},
  {"04426LIST", 1},
  {"04451LIST", 1},
  {"04450LIST", 1},
  {"04500LIST", 1},
  {"04501LIST", 1},
  {"04502LIST", 1},
  {"04421LIST", 1},
  {"04530LIST", 1},
  {"04534LIST", 1},
  {"04535LIST", 1},

  //ftp_answer_hash_add ("04LPRT");

  //ftp_answer_hash_add ("04LPSV");

  {"04151MAIL", 1},
  {"04152MAIL", 1},
  {"04354MAIL", 1},
  {"04250MAIL", 1},
  {"04451MAIL", 1},
  {"04552MAIL", 1},
  {"04450MAIL", 1},
  {"04550MAIL", 1},
  {"04452MAIL", 1},
  {"04553MAIL", 1},
  {"04500MAIL", 1},
  {"04501MAIL", 1},
  {"04502MAIL", 1},
  {"04421MAIL", 1},
  {"04530MAIL", 1},

  //ftp_answer_hash_add ("04MDTM");

  {"04535MIC", 1},
  {"04533MIC", 1},
  {"04500MIC", 1},
  {"04501MIC", 1},
  {"04421MIC", 1},

  {"04257MKD", 1},
  {"04500MKD", 1},
  {"04501MKD", 1},
  {"04502MKD", 1},
  {"04421MKD", 1},
  {"04530MKD", 1},
  {"04550MKD", 1},

  {"04125MLFL", 1},
  {"04150MLFL", 1},
  {"04226MLFL", 1},
  {"04250MLFL", 1},
  {"04425MLFL", 1},
  {"04426MLFL", 1},
  {"04451MLFL", 1},
  {"04552MLFL", 1},
  {"04532MLFL", 1},
  {"04450MLFL", 1},
  {"04550MLFL", 1},
  {"04452MLFL", 1},
  {"04553MLFL", 1},
  {"04500MLFL", 1},
  {"04501MLFL", 1},
  {"04502MLFL", 1},
  {"04421MLFL", 1},
  {"04530MLFL", 1},

  {"04200MODE", 1},
  {"04500MODE", 1},
  {"04501MODE", 1},
  {"04504MODE", 1},
  {"04421MODE", 1},
  {"04530MODE", 1},

  {"04151MRCP", 1},
  {"04152MRCP", 1},
  {"04200MRCP", 1},
  {"04450MRCP", 1},
  {"04550MRCP", 1},
  {"04452MRCP", 1},
  {"04553MRCP", 1},
  {"04500MRCP", 1},
  {"04501MRCP", 1},
  {"04502MRCP", 1},
  {"04503MRCP", 1},
  {"04421MRCP", 1},

  {"04200MRSQ", 1},
  {"04215MRSQ", 1},
  {"04500MRSQ", 1},
  {"04501MRSQ", 1},
  {"04502MRSQ", 1},
  {"04421MRSQ", 1},
  {"04530MRSQ", 1},

  {"04119MSAM", 1},
  {"04151MSAM", 1},
  {"04152MSAM", 1},
  {"04354MSAM", 1},
  {"04250MSAM", 1},
  {"04451MSAM", 1},
  {"04552MSAM", 1},
  {"04450MSAM", 1},
  {"04550MSAM", 1},
  {"04452MSAM", 1},
  {"04553MSAM", 1},
  {"04500MSAM", 1},
  {"04501MSAM", 1},
  {"04502MSAM", 1},
  {"04421MSAM", 1},
  {"04530MSAM", 1},

  {"04151MSND", 1},
  {"04152MSND", 1},
  {"04354MSND", 1},
  {"04250MSND", 1},
  {"04451MSND", 1},
  {"04552MSND", 1},
  {"04450MSND", 1},
  {"04550MSND", 1},
  {"04452MSND", 1},
  {"04553MSND", 1},
  {"04500MSND", 1},
  {"04501MSND", 1},
  {"04502MSND", 1},
  {"04421MSND", 1},
  {"04530MSND", 1},

  {"04119MSOM", 1},
  {"04151MSOM", 1},
  {"04152MSOM", 1},
  {"04354MSOM", 1},
  {"04250MSOM", 1},
  {"04451MSOM", 1},
  {"04552MSOM", 1},
  {"04450MSOM", 1},
  {"04550MSOM", 1},
  {"04452MSOM", 1},
  {"04553MSOM", 1},
  {"04500MSOM", 1},
  {"04501MSOM", 1},
  {"04502MSOM", 1},
  {"04421MSOM", 1},
  {"04530MSOM", 1},

  {"04125NLST", 1},
  {"04150NLST", 1},
  {"04226NLST", 1},
  {"04250NLST", 1},
  {"04425NLST", 1},
  {"04426NLST", 1},
  {"04451NLST", 1},
  {"04450NLST", 1},
  {"04500NLST", 1},
  {"04501NLST", 1},
  {"04502NLST", 1},
  {"04421NLST", 1},
  {"04530NLST", 1},
  {"04534NLST", 1},
  {"04535NLST", 1},
  // WuFTPd
  {"04550NLST", 1},

  {"04200NOOP", 1},
  {"04500NOOP", 1},
  {"04421NOOP", 1},

  //ftp_answer_hash_add ("04OPTS");

  {"04230PASS", 1},
  {"04202PASS", 1},
  {"04530PASS", 1},
  {"04500PASS", 1},
  {"04501PASS", 1},
  {"04503PASS", 1},
  {"04421PASS", 1},
  {"04332PASS", 1},

// in conversion mode
  {"04200PASV", 1},

  {"04227PASV", 1},
  {"04500PASV", 1},
  {"04501PASV", 1},
  {"04502PASV", 1},
  {"04421PASV", 1},
  {"04530PASV", 1},

  {"04200PBSZ", 1},
  {"04503PBSZ", 1},
  {"04500PBSZ", 1},
  {"04501PBSZ", 1},
  {"04421PBSZ", 1},
  {"04530PBSZ", 1},

  {"04200PORT", 1},

// in conversion mode
  {"04227PORT", 1},
  
  {"04500PORT", 1},
  {"04501PORT", 1},
  {"04421PORT", 1},
  {"04530PORT", 1},

  {"04200PROT", 1},
  {"04504PROT", 1},
  {"04536PROT", 1},
  {"04503PROT", 1},
  {"04534PROT", 1},
  {"04431PROT", 1},
  {"04500PROT", 1},
  {"04501PROT", 1},
  {"04421PROT", 1},
  {"04530PROT", 1},

  {"04257PWD", 1},
  {"04500PWD", 1},
  {"04501PWD", 1},
  {"04502PWD", 1},
  {"04421PWD", 1},
  {"04550PWD", 1},

  {"04221QUIT", 1},
  {"04500QUIT", 1},

  {"04120REIN", 1},
  {"04220REIN", 1},
  {"04421REIN", 1},
  {"04500REIN", 1},
  {"04502REIN", 1},

  {"04500REST", 1},
  {"04501REST", 1},
  {"04502REST", 1},
  {"04421REST", 1},
  {"04530REST", 1},
  {"04350REST", 1},

  {"04125RETR", 1},
  {"04150RETR", 1},
  {"04110RETR", 1},
  {"04226RETR", 1},
  {"04250RETR", 1},
  {"04425RETR", 1},
  {"04426RETR", 1},
  {"04451RETR", 1},
  {"04450RETR", 1},
  {"04550RETR", 1},
  {"04500RETR", 1},
  {"04501RETR", 1},
  {"04421RETR", 1},
  {"04530RETR", 1},
  {"04534RETR", 1},
  {"04535RETR", 1},

  {"04250RMD", 1},
  {"04500RMD", 1},
  {"04501RMD", 1},
  {"04502RMD", 1},
  {"04421RMD", 1},
  {"04530RMD", 1},
  {"04550RMD", 1},

  {"04450RNFR", 1},
  {"04550RNFR", 1},
  {"04500RNFR", 1},
  {"04501RNFR", 1},
  {"04502RNFR", 1},
  {"04421RNFR", 1},
  {"04530RNFR", 1},
  {"04350RNFR", 1},

  {"04250RNTO", 1},
  {"04532RNTO", 1},
  {"04553RNTO", 1},
  {"04500RNTO", 1},
  {"04501RNTO", 1},
  {"04502RNTO", 1},
  {"04503RNTO", 1},
  {"04421RNTO", 1},
  {"04530RNTO", 1},

  {"04200SITE", 1},
  {"04202SITE", 1},
  {"04500SITE", 1},
  {"04501SITE", 1},
  {"04530SITE", 1},

  // NcFTP
  {"04213SIZE", 1},
  {"04500SIZE", 1},

  {"04202SMNT", 1},
  {"04250SMNT", 1},
  {"04500SMNT", 1},
  {"04501SMNT", 1},
  {"04502SMNT", 1},
  {"04421SMNT", 1},
  {"04530SMNT", 1},
  {"04550SMNT", 1},

  {"04211STAT", 1},
  {"04212STAT", 1},
  {"04213STAT", 1},
  {"04450STAT", 1},
  {"04500STAT", 1},
  {"04501STAT", 1},
  {"04502STAT", 1},
  {"04421STAT", 1},
  {"04530STAT", 1},

  {"04125STOR", 1},
  {"04150STOR", 1},
  {"04110STOR", 1},
  {"04226STOR", 1},
  {"04250STOR", 1},
  {"04425STOR", 1},
  {"04426STOR", 1},
  {"04451STOR", 1},
  {"04551STOR", 1},
  {"04552STOR", 1},
  {"04532STOR", 1},
  {"04450STOR", 1},
  {"04452STOR", 1},
  {"04553STOR", 1},
  {"04500STOR", 1},
  {"04501STOR", 1},
  {"04421STOR", 1},
  {"04530STOR", 1},
  {"04534STOR", 1},
  {"04535STOR", 1},

  {"04125STOU", 1},
  {"04150STOU", 1},
  {"04110STOU", 1},
  {"04226STOU", 1},
  {"04250STOU", 1},
  {"04425STOU", 1},
  {"04426STOU", 1},
  {"04451STOU", 1},
  {"04551STOU", 1},
  {"04552STOU", 1},
  {"04532STOU", 1},
  {"04450STOU", 1},
  {"04452STOU", 1},
  {"04553STOU", 1},
  {"04500STOU", 1},
  {"04501STOU", 1},
  {"04421STOU", 1},
  {"04530STOU", 1},
  {"04534STOU", 1},
  {"04535STOU", 1},

  {"04200STRU", 1},
  {"04500STRU", 1},
  {"04501STRU", 1},
  {"04504STRU", 1},
  {"04421STRU", 1},
  {"04530STRU", 1},

  {"04215SYST", 1},
  {"04500SYST", 1},
  {"04501SYST", 1},
  {"04502SYST", 1},
  {"04421SYST", 1},

  {"04200TYPE", 1},
  {"04500TYPE", 1},
  {"04501TYPE", 1},
  {"04504TYPE", 1},
  {"04421TYPE", 1},
  {"04530TYPE", 1},

  {"04230USER", 1},
  {"04232USER", 1},
  {"04530USER", 1},
  {"04500USER", 1},
  {"04501USER", 1},
  {"04421USER", 1},
  {"04331USER", 1},
  {"04332USER", 1},
  {"04336USER", 1},

  {"04200XCUP", 1},
  {"04506XCUP", 1},
  {"04502XCUP", 1},
  {"04507XCUP", 1},
  {"04551XCUP", 1},
  {"04521XCUP", 1},
  {"04450XCUP", 1},
  {"04425XCUP", 1},
  {"04451XCUP", 1},

  //ftp_answer_hash_add ("04XCWD");

  {"04257XMKD", 1},
  {"04251XMKD", 1},
  {"04521XMKD", 1},
  {"04450XMKD", 1},
  {"04506XMKD", 1},
  {"04502XMKD", 1},
  {"04550XMKD", 1},
  {"04501XMKD", 1},
  {"04425XMKD", 1},
  {"04451XMKD", 1},

  {"04257XPWD", 1},
  {"04251XPWD", 1},
  {"04425XPWD", 1},
  {"04451XPWD", 1},
  {"04506XPWD", 1},
  {"04502XPWD", 1},

  {"04224XRMD", 1},
  {"04250XRMD", 1},
  {"04506XRMD", 1},
  {"04502XRMD", 1},
  {"04521XRMD", 1},
  {"04450XRMD", 1},
  {"04550XRMD", 1},
  {"04501XRMD", 1},
  {"04425XRMD", 1},
  {"04451XRMD", 1},

  // NcFTP
  //{"04500CLNT", 1},

  {NULL, 0}
};

// Function:    ftp_disable_answer
// In:          ZProxy * s              object pointer
//              ZPolicyObj * args       proxy args
// Out:         ZPolicyObj *
// Desc:        disable an answer
ZPolicyObj *
ftp_disable_answer (ZProxy * s, ZPolicyObj * args)
{
  FtpProxy *self = (FtpProxy *) s;
  char *strname;
  char str[2 + 3 + 4 + 1];
  FTPAnswerDescriptor *wp = NULL;
  //int i = 0;

#if ENABLE_DEBUG
  z_debug (7, "%s: Disable answer", self->super.session_id);
#endif

  if (z_policy_var_parse (args, "(s)", &strname))
    {
      snprintf (str, 10, "04%s", strname);
#if ENABLE_DEBUG
      z_debug (7, "%s: Answer -%s- ", self->super.session_id, str);
#endif
      if ((wp = g_hash_table_lookup (self->local_answer, str)) != NULL)
	{
	  g_hash_table_remove (self->local_answer, str);
	  g_free (wp->info);
	  g_free (wp);
	}
    }

  z_policy_var_ref (z_policy_none);
  return z_policy_none;
  self = NULL;
}

// Function:    ftp_enable_answer
// In:          ZProxy * s              object pointer
//              ZPolicyObj * args       proxy args
// Out:         ZPolicyObj *
// Desc:        enable an answer
ZPolicyObj *
ftp_enable_answer (ZProxy * s, ZPolicyObj * args)
{
  FtpProxy *self = (FtpProxy *) s;
  char *strcode;
  char *strname;
  char str[2 + 3 + 4 + 1];
  //char *tmp;
  FTPAnswerDescriptor *wp = NULL;
  //int i = 0;

#if ENABLE_DEBUG
  z_debug (7, "%s: Enable answer", self->super.session_id);
#endif

  if (z_policy_var_parse (args, "(ss)", &strcode, &strname))
    {
      snprintf (str, 10, "04%3s%s", strcode, strname);
#if ENABLE_DEBUG
      z_debug (7, "%s: Answer -%s- ", self->super.session_id, str);
#endif
      if ((wp = g_hash_table_lookup (self->local_command, strname)) == NULL)
	{
	  wp = g_new0 (FTPAnswerDescriptor, 1);
	  wp->info = strdup (str);
	  wp->number = 1;
	  g_hash_table_insert (self->local_answer, wp->info, wp);
#if ENABLE_DEBUG
	  z_debug (7, "%s: local_answer size %d ", self->super.session_id, g_hash_table_size (self->local_answer));
	  if (wp)
	    z_debug (7, "%s: local_answer -%s- %d ", self->super.session_id, wp->info, wp->number);
#endif
	}
    }
  else
    {
#if ENABLE_DEBUG
  z_debug (7, "%s: Enable answer: failed var_parse", self->super.session_id);
#endif
    }

  z_policy_var_ref (z_policy_none);
  return z_policy_none;
  self = NULL;
}

// Function:    ftp_disable_command
// In:          ZProxy * s              object pointer
//              ZPolicyObj * args       proxy args
// Out:         ZPolicyObj *
// Desc:        disable a command
ZPolicyObj *
ftp_disable_command (ZProxy * s, ZPolicyObj * args)
{
  FtpProxy *self = (FtpProxy *) s;
  char *strname;
  FTPInternalCommand *wp = NULL;

#if ENABLE_DEBUG
  z_debug (7, "%s: Disable command", self->super.session_id);
#endif

  if (z_policy_var_parse (args, "(s)", &strname))
    {
#if ENABLE_DEBUG
      z_debug (7, "%s: Command -%s- ", self->super.session_id, strname);
#endif
      if ((wp = g_hash_table_lookup (self->local_command, strname)) != NULL)
	{
	  wp->number = 0;
	  g_hash_table_insert (self->local_command, wp->info, wp);
	}
    }

  z_policy_var_ref (z_policy_none);
  return z_policy_none;
  self = NULL;
}

// Function:    ftp_enable_command
// In:          ZProxy * s              object pointer
//              ZPolicyObj * args       proxy args
// Out:         ZPolicyObj *
// Desc:        enable a command
ZPolicyObj *
ftp_enable_command (ZProxy * s, ZPolicyObj * args)
{
  FtpProxy *self = (FtpProxy *) s;
  char *strname;
  FTPInternalCommand *wp = NULL;

#if ENABLE_DEBUG
  z_debug (7, "%s: Enable command", self->super.session_id);
#endif

  if (z_policy_var_parse (args, "(s)", &strname))
    {
#if ENABLE_DEBUG
      z_debug (7, "%s: Command -%s- ", self->super.session_id, strname);
#endif
      if ((wp = g_hash_table_lookup (self->local_command, strname)) != NULL)
	{
	  wp->number = 1;
	  g_hash_table_insert (self->local_command, wp->info, wp);
	}
    }

  z_policy_var_ref (z_policy_none);
  return z_policy_none;
  self = NULL;
}

// Function:    ftp_destroy
// In:          ZProxy * s              object pointer
//              ZPolicyObj * args       proxy args
// Out:         ZPolicyObj *
// Desc:        test callback function, called from the core
ZPolicyObj *
ftp_destroy (ZProxy * s, ZPolicyObj * args)
{
  FtpProxy *self = (FtpProxy *) s;
  char *strname;

  z_debug (7, "%s: In the callback...", self->super.session_id);
  if (z_policy_var_parse (args, "(s)", &strname))
    {
      z_debug (7, "%s: FTP CallBack -%s- ", self->super.session_id, strname);
    }

  z_debug (7, "%s: FTP CallBack ", self->super.session_id);

  z_policy_var_ref (z_policy_none);
  return z_policy_none;
  self = NULL;
}

// Function:    ftp_local_answer_hash_get
// In:          FtpProxy*               proxy pointer
//              gchar                   command
// Out:         int                     value                       
// Desc:        look up the answer in the local_answer
int
ftp_local_answer_hash_get (FtpProxy * self, gchar * info)
{
  FTPAnswerDescriptor *wp = g_hash_table_lookup (self->local_answer, info);

#if ENABLE_DEBUG
  z_debug (7, "%s: local_answer size %d ", self->super.session_id, g_hash_table_size (self->local_answer));
  if (wp)
    z_debug (7, "%s: local_answer -%s- %d ", self->super.session_id, wp->info, wp->number);
#endif

  if (wp)
    return wp->number;
  else
    return 0;
}

// Function:    ftp_local_hash_get
// In:          FtpProxy*               proxy pointer
//              gchar                   command
// Out:         int                     value                       
// Desc:        look up the command in the local_hash
int
ftp_local_hash_get (FtpProxy * self, gchar * info)
{
  FTPInternalCommand *wp = g_hash_table_lookup (self->local_command, info);
  if (wp)
    return wp->number;
  else
    return 0;
}

// Function:    ftp_answer_hash_get
// In:          gchar                   answer
// Out:         int                     value                       
// Desc:        look up the answer in the ftp_answer_hash
int
ftp_answer_hash_get (gchar * info)
{
  FTPAnswerDescriptor *wp = g_hash_table_lookup (ftp_answer_hash, info);
  if (wp)
    return wp->number;
  else
    return 0;
}

// Function:    ftp_command_hash_get
// In:          gchar                   command
// Out:         gpointer                value                       
// Desc:        look up the command in the ftp_command_hash
//              return: a pointer to the structure
gpointer ftp_command_hash_get (gchar * name)
{
  FTPCommandDescriptor *wp = g_hash_table_lookup (ftp_command_hash, name);
  return wp;
}

// Function:    ftp_command_hash_get
// In:          gchar                   command
// Out:         gpointer                value                       
// Desc:        look up the command in the ftp_command_hash
//              return: a pointer to the function in the structure
gpointer ftp_command_hash_get_fv (gchar * name)
{
  FTPCommandDescriptor *wp = g_hash_table_lookup (ftp_command_hash, name);
  if (wp)
    return wp->fv;
  else
    return NULL;
}

// Function:    ftp_proxy_regvars
// In:          FtpProxy*               proxy pointer
// Out:         -                       
// Desc:        register variables for Python interface
void
ftp_proxy_regvars (FtpProxy * self)
{
  z_proxy_var_new (&self->super, "proxy_state", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->proxy_state);
  z_proxy_var_new (&self->super, "client_portpasv", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->client_portpasv);
  z_proxy_var_new (&self->super, "internal_server", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->internal_server);
  z_proxy_var_new (&self->super, "NAT", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->NAT);
  z_proxy_var_new (&self->super, "verbose_messages", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->verbose_messages);
  z_proxy_var_new (&self->super, "bounce_check", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->verbose_messages);
  z_proxy_var_new (&self->super, "max_line_length", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->max_line_length);
  z_proxy_var_new (&self->super, "max_username_length", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->max_username_length);
  z_proxy_var_new (&self->super, "max_password_length", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->max_password_length);
  z_proxy_var_new (&self->super, "max_pathname_length", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->max_pathname_length);
  z_proxy_var_new (&self->super, "client_data_address", Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET_CONFIG, self->client_data_address);
  z_proxy_var_new (&self->super, "server_data_address", Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, self->server_data_address);
  z_proxy_var_new (&self->super, "fw_internal_data_address", Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, self->fw_internal_data_address);
  z_proxy_var_new (&self->super, "fw_external_data_address", Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, self->fw_external_data_address);
  z_proxy_var_new (&self->super, "tmp_data_address", Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, self->tmp_data_address);
  z_proxy_var_new (&self->super, "client_data_port", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->client_data_port);
  z_proxy_var_new (&self->super, "server_data_port", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->server_data_port);
  z_proxy_var_new (&self->super, "fw_internal_data_port", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->fw_internal_data_port);
  z_proxy_var_new (&self->super, "fw_external_data_port", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->fw_external_data_port);
  z_proxy_var_new (&self->super, "tmp_data_port", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->tmp_data_port);
  z_proxy_var_new (&self->super, "path_separator", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->path_separator);
  z_proxy_var_new (&self->super, "command_accept", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->command_accept);
  z_proxy_var_new (&self->super, "command_log", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->command_log);
  z_proxy_var_new (&self->super, "command_up", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->command_up);
  z_proxy_var_new (&self->super, "allow_PWD", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_SET, &self->allow_PWD);
  z_proxy_var_new (&self->super, "FtpAnswerENABLE", Z_VAR_GET | Z_VAR_TYPE_METHOD, ftp_enable_answer);
  z_proxy_var_new (&self->super, "FtpAnswerDISABLE", Z_VAR_GET | Z_VAR_TYPE_METHOD, ftp_disable_answer);
  z_proxy_var_new (&self->super, "FtpCommandENABLE", Z_VAR_GET | Z_VAR_TYPE_METHOD, ftp_enable_command);
  z_proxy_var_new (&self->super, "FtpCommandDISABLE", Z_VAR_GET | Z_VAR_TYPE_METHOD, ftp_disable_command);
  z_proxy_var_new (&self->super, "ftpDestroy", Z_VAR_GET | Z_VAR_TYPE_METHOD, ftp_destroy);

}

// Function:    ftp_create_hash
// In:          FtpProxy*               proxy pointer
// Out:         -                       
// Desc:        create and fill the hashtable
void
ftp_create_hash (FtpProxy * self)
{
  int i;

  ftp_command_hash = g_hash_table_new (g_str_hash, g_str_equal);

  i = 0;

  while (ftp_commands[i].name != NULL)
    {
      if (g_hash_table_lookup (ftp_command_hash, ftp_commands[i].name) == NULL)
	{
	  g_hash_table_insert (ftp_command_hash, ftp_commands[i].name, &ftp_commands[i]);
	}
      i++;
    }

}


// Function:    ftp_create_local_hash
// In:          FtpProxy*               proxy pointer
// Out:         -                       
// Desc:        create and fill the hashtable
void
ftp_create_local_hash (FtpProxy * self)
{
  int i;

  self->local_command = g_hash_table_new (g_str_hash, g_str_equal);
  self->local_answer = g_hash_table_new (g_str_hash, g_str_equal);

  i = 0;

  while (ftp_internal_commands[i].info != NULL)
    {
      if (g_hash_table_lookup (self->local_command, ftp_internal_commands[i].info) == NULL)
	{
	  g_hash_table_insert (self->local_command, ftp_internal_commands[i].info, &ftp_internal_commands[i]);
	}
      i++;
    }
}

// Function:    ftp_create_answer_hash
// In:          FtpProxy*               proxy pointer
// Out:         -                       
// Desc:        create and fill the hashtable
void
ftp_create_answer_hash (FtpProxy * self)
{
  int i;

  ftp_answer_hash = g_hash_table_new (g_str_hash, g_str_equal);

  i = 0;

  while (ftp_answers[i].info != NULL)
    {
      if (g_hash_table_lookup (ftp_answer_hash, ftp_answers[i].info) == NULL)
	{
	  g_hash_table_insert (ftp_answer_hash, ftp_answers[i].info, &ftp_answers[i]);
	}
#ifdef ENABLE_DEBUG
      else
	{
	  z_debug (7, "%s: FTP: ftp_answer_hash %d %s", self->super.session_id, i, ftp_answers[i].info);
	}
#endif
      i++;
    }

}

// Function:    ftp_create_conf
// In:          FtpProxy*               proxy pointer
// Out:         int                     0 - failed                       
// Desc:        fill the variables...
int
ftp_create_conf (FtpProxy * self)
{

  self->ftp_state = FTP_STATE_CONNECT;
  self->proxy_state = FALSE;
  self->client_portpasv = 0;
  self->NAT = FALSE;
  self->bounce_check = TRUE;
  self->verbose_messages = TRUE;
  self->max_line_length = 255;
  self->max_username_length = 8;
  self->max_password_length = 8;
  self->max_pathname_length = 255;
  self->username = g_new0 (char, self->max_username_length + 1);
  self->password = g_new0 (char, self->max_password_length + 1);
  self->parameter = g_new0 (char, self->max_line_length + 1);
  self->path = g_new0 (char, self->max_pathname_length);
  self->path_separator = '/';
  self->allow_PWD = 1;
  self->command_accept = 0x00000000L;
  self->command_log = 0xFFFFFFFFL;
  self->command_unk = 0;

  self->server_answer_begin = TRUE;
  self->log_dropped = TRUE;
  self->log_answer = TRUE;

  self->server_data_address = g_string_new ("0.0.0.0");
  self->client_data_address = g_string_new ("0.0.0.0");

  self->fw_external_data_address = g_string_new ("0.0.0.0");
  self->fw_internal_data_address = g_string_new ("0.0.0.0");

  self->tmp_data_address = g_string_new ("0.0.0.0");

  self->server_data_port = 20;
  self->client_data_port = 0;
  self->fw_external_data_port = 0;
  self->fw_internal_data_port = 0;
  self->tmp_data_port = 0;
  self->last_port = 2;
  self->tmp_port = 2;
  self->internal_server = 1;

  ftp_create_local_hash (self);

  return 1;
}
