/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: httpfltr.c,v 1.11 2000/08/08 14:47:17 bazsi Exp $
 *
 * By: Kovacs Viktor Peter (vps__@hotmail.com | vps__@freemail.hu | ...)
 *
 ***************************************************************************/
#include "http.h"

//#define TRUE  (1)
//#define FALSE (0)

int http_transfer_filter_copy(
 HttpProxy* s, 
 int side_src,
 int side_dst,
 int (*content_enc)(HttpProxy* s, HttpStr* buf, int (*content_typ)(HttpProxy* s, HttpStr* buf, int len, int* errorval), int len, int* errorval),
 int (*content_typ)(HttpProxy* s, HttpStr* buf, int len, int* errorval),
 int len,
 int* errorval)
{
 int co=0, t;
 int srv=TRUE;
 HttpStr  tsd;
 HttpStr* ts=&tsd;
 http_str_init(ts);

 *errorval=0;
 if(!((side_src==HTTP_SIDE_CLIENT && side_dst==HTTP_SIDE_SERVER) || (side_src==HTTP_SIDE_SERVER && side_dst==HTTP_SIDE_CLIENT)))
 {
  *errorval=HTTP_SIDE_BOTH;
  http_str_free(ts); 
  return(HTTP_ERROR);
 }
 
 http_str_sets(ts, ""); //terminate headers with crlf
 if(http_writeline(s, side_dst, ts)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); return(HTTP_ERROR);}

 http_str_allocfill(ts, HTTP_BLOCKSIZE, 0);
 if(len!=-1)
 {
  while(1==1)
  {
   if(srv==TRUE)
   {
    //ts->len=http_less(HTTP_BLOCKSIZE, len-co);
    if((t=http_read(s, side_src, ts))==HTTP_ERROR){*errorval=side_src; http_str_free(ts); return(HTTP_ERROR);}
    co+=t;
    if(co>=len) srv=FALSE;
    //if(co>len){*errorval=side_src; http_str_free(ts); return(HTTP_ERROR);} //=> problem with url encoded fields
   }else
   {
    t=0;
   }
   t=content_enc(s, ts, content_typ, t, errorval);
   if(t>0)
   {
    if(http_send(s, side_dst, (ts->str), t)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); return(HTTP_ERROR);}
   }
   if(t==0){http_str_free(ts); return(0);}else        //ok (transfer complete)
   if(t==-1){http_str_free(ts); return(HTTP_ERROR);}  //stacked filter error
  }
 }else if(len==-1)
 {
  while(1==1)
  {
   if(srv==TRUE)
   {
    if((t=http_read(s, side_src, ts))==HTTP_ERROR){*errorval=side_src; http_str_free(ts); return(HTTP_ERROR);}
    if(t==0){srv=FALSE;}
   }else
   {
    t=0;
   }
   t=content_enc(s, ts, content_typ, t, errorval);
   //z_debug(0, "transfer (%d)", t);
   if(t>0)
   {
    if(http_send(s, side_dst, ts->str, t)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); return(HTTP_ERROR);}
   }
   if(t==0){http_str_free(ts); return(0);}else        //ok (transfer complete)
   if(t==-1){http_str_free(ts); return(HTTP_ERROR);}  //stacked filter error
  } 
 }
 http_str_free(ts);
 return(0);
}

int http_encoding_filter_binary(
 HttpProxy* s,
 HttpStr* buf,
 int (*content_typ)(HttpProxy* s, HttpStr* buf, int len, int* errorval),
 int len,
 int* errorval)
{
 return(content_typ(s, buf, len, errorval));
}

int http_type_filter_binary(
 HttpProxy* s,
 HttpStr* buf,
 int len,
 int* errorval)
{
 return(len);
}

//----------------------------------------------------------------
int http_transfer_filter_chunked(
 HttpProxy* s, 
 int side_src,
 int side_dst,
 int (*content_enc)(HttpProxy* s, HttpStr* buf, int (*content_typ)(HttpProxy* s, HttpStr* buf, int len, int* errorval), int len, int* errorval),
 int (*content_typ)(HttpProxy* s, HttpStr* buf, int len, int* errorval),
 int len,
 int* errorval)
{
 int co=0, clen=0, t;
 int srv=TRUE;
 HttpStr  tsd;
 HttpStr* ts=&tsd;
 HttpStr  hsd;
 HttpStr* hs=&hsd;
 http_str_init(ts);
 http_str_init(hs);

 *errorval=0;
 if(!((side_src==HTTP_SIDE_CLIENT && side_dst==HTTP_SIDE_SERVER) || (side_src==HTTP_SIDE_SERVER && side_dst==HTTP_SIDE_CLIENT)))
 {
  *errorval=HTTP_SIDE_BOTH;
  http_str_free(ts); http_str_free(hs);
  return(HTTP_ERROR);
 }
 
 http_str_sets(hs, ""); //terminate headers with crlf
 if(http_writeline(s, side_dst, hs)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}

 http_str_allocfill(ts, HTTP_BLOCKSIZE, 0);
 srv=TRUE;
 co=0; clen=0;
 while(1==1)
 {
  if(srv==TRUE)
  {
   if(co==clen) //get new block data
   {
    co=0;
    if(http_readline(s, side_src, hs, s->maxlinelen)==HTTP_ERROR){*errorval=side_src; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}
    clen=0; //use strtol instead of sscanf!!!
    if(sscanf(hs->str, "%x", &clen)!=1){*errorval=side_src; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}
    if(clen==0) srv=FALSE; //no more blocks
   }
   if(co<clen) //read data
   {
    ts->len=http_less(HTTP_BLOCKSIZE, clen-co);
    if((t=http_read(s, side_src, ts))==HTTP_ERROR){*errorval=side_src; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}
    co+=t;
    if(co==clen) //get trailing crlf
     if(http_readline(s, side_src, hs, s->maxlinelen)==HTTP_ERROR){*errorval=side_src; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}
   }
  }else //no more data from the server
  {
   t=0;
  }
  
  t=content_enc(s, ts, content_typ, t, errorval); //filter
  if(t>0) //write data
  {
   http_str_allocfill(ts, s->maxlinelen, 0); //alloc block header
   g_snprintf(hs->str, s->maxlinelen, "%x", t);
   //send data
   if(http_writeline(s, side_dst, hs)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}   
   if(http_send(s, side_dst, (ts->str), t)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}
   http_str_sets(hs, ""); //send terminator (crlf)
   if(http_writeline(s, side_dst, hs)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}   
  }
  if(t==-1){http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);} //error in stacked filter
  if(t==0) //end of transfer
  {
   http_str_sets(hs, "0"); //send terminator block
   if(http_writeline(s, side_dst, hs)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}   
   break;
  }
 }
 
 http_str_sets(hs, ""); 
 do //copy footers: needs filtering
 {
  if(http_readline(s, side_src, hs, s->maxlinelen)==HTTP_ERROR){*errorval=side_src; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}
  //call header filter!!! (if hs->len!=0)
  if(http_writeline(s, side_dst, hs)==HTTP_ERROR){*errorval=side_dst; http_str_free(ts); http_str_free(hs); return(HTTP_ERROR);}
 }while(hs->len!=0);
 
 http_str_free(ts); http_str_free(hs);
 return(0);
}
