############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: MacPolicy.py,v 1.2 2000/08/01 10:10:37 bazsi Exp $
##
############################################################################

MACException = 'MACException'

class MacPolicy:
	def isOperationPermitted(self, session, dir):
		# this should return one of Z_ACCEPT, Z_REJECT
		raise NotImplementedError

class StrictMAC:
	"""
	   Mix-in class to be included in event handlers to provide strict 
	   MAC policy checking.
	"""
	def preProcessEvent(self, dir):
		""" The preProcessEvent is called just before the administrator
		supplied event handler. If it returns Z_UNSPEC, the event
		handler of the administrator is called, if it returns 
		Z_ACCEPT, Z_DENY or Z_REJECT the user supplied event handler
		is not even called. """

		res = self.mac.isOperationPermitted(self, dir)
		if (res == Z_ACCEPT):
			# give a chance to DAC
			return Z_UNSPEC
		return res
		
class LooseMAC:
	"""
	   Mix-in class to be included in event handlers to provide loose
	   MAC policy checking. This means that the administrator can 
	   override MAC policy decisions.
	"""
	def postProcessEvent(self, dir):
		""" The postProcessEvent is called if neither preProcessEvent
		nor the administrator supplied event handler returned a suitable
		action code."""
		
		return self.mac.isOperationPermitted(self, dir)

