############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Service.py,v 1.9 2000/08/01 10:10:37 bazsi Exp $
##
############################################################################

import Zorp, Session, Stream

#class ServiceInstance:
#	client = None
#	zone = None
#	
#	def __init__(self, zone, client, param):
#		self.zone = zone
#		self.client = client

class AbstractService:
	"""
		Service kontainer osztlya

		Vltozk:
			name			A service neve
			instance_id	A service ltal kiszolglt session
	"""
	name = ""
	instance_id = 0
	
	def __init__(self, name):
		"""
			Ltrehoz egy AbstractService pldnyt.

			Vltozk:
				name		A service neve

			Lers:
				-

			Visszatrsi rtk:
				-
		"""
		self.name = name
	
	def startInstance(self, session):
		raise NotImplementedError

class Service(AbstractService):
	"""
		Szolgltats osztly feladata a szolgltatshoz tartoz kiegszt
		szolgltatsok elindtsa (esemnykezel)

		Vltozk:
			chainer			A szolgltatshoz tartoz chainer osztly
			proxy_class		A szolgltatshoz tartoz proxy osztlya
	"""
	chainer = None
	proxy_class = None
	auth = None

	def __init__(self, name, chainer, proxy_class, auth = None):
		"""
			Ltrehozza az osztly egy pldnyt.

			Paramterek:
				chainer		A szolgltatshoz tartoz chainer osztly
				proxy_class	A szolgltatshoz tartoz proxy osztlya

			Lers:
				-

			Visszatrsi rtk:
				-
		"""
		AbstractService.__init__(self, name)
		self.chainer = chainer
		self.proxy_class = proxy_class
		self.auth = auth

	def startInstance(self, session):
		"""
			Callback fggvny. A Listener ltal elkapott kapcsolatok
			kezelst vgzi.

			Paramterek:
				session		A kapcsolat lerja (Session.session)

			Lers: 
				Azonostja a felhasznlt, majd ltrehozza a
				kapcsolathoz tartoz callback
				fggvnyosztly egy pldnyt

			Visszatrsi rtk:
				-
		"""

		if session.auth == None:
			session.auth = self.auth		
		if session.auth != None and not session.auth.authenticate(session):
			raise Zorp.AuthException

		Zorp.message(0, "auth_user=" + session.auth_user)
		
		session.name = self.name
		instance_id = self.instance_id
		
		# FIXME: handle wraparound
		self.instance_id = self.instance_id + 1
		
		session.setServiceInstance(instance_id)
		ss = Session.StackedSession(session, self.chainer)
		ss.client = Stream.Stream(ss.client_fd, ss.session_id + ": client")
		self.proxy_class(ss)
			
		return None
		
#
# Egy instance_class es param helyett egy mappingot vesz at, ami zonatol
# fuggoen mas es mas eredmenyt ad
#
#class MultiService(AbstractService):
#	def __init__(self, name
